#define PROGRAM_FILE "add_vectors.cl"
#define KERNEL_FUNC "add_vectors"

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#ifdef MAC
#include <OpenCL/cl.h>
#else
#include <CL/cl.h>
#endif

cl_device_id create_device() {

   cl_platform_id platform;
   cl_device_id dev;
   int err;

   /* Identify a platform */
   err = clGetPlatformIDs(1, &platform, NULL);
   if(err < 0) {
      perror("Couldn't identify a platform");
      exit(1);
   } 

   // Access a device
   // GPU
   err = clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, 1, &dev, NULL);
   if(err == CL_DEVICE_NOT_FOUND) {
      // CPU
      err = clGetDeviceIDs(platform, CL_DEVICE_TYPE_CPU, 1, &dev, NULL);
   }
   if(err < 0) {
      perror("Couldn't access any devices");
      exit(1);   
   }

   return dev;
}

cl_context create_context(cl_device_id device) 
{
  cl_context context;
  int err;

  context = clCreateContext(NULL, 1, &device, NULL, NULL, &err);
  if(err < 0) {
    perror("Couldn't create a context");
    exit(1);   
  }
  return context;
}

cl_command_queue create_queue(cl_device_id device, cl_context context)
{
  int err;
  cl_command_queue queue = clCreateCommandQueueWithProperties(
                context, device, 0, &err);
  if(err < 0) {
    perror("Couldn't create a command queue");
    exit(1);   
  };
  return queue;
}

cl_program build_program(cl_context ctx, 
                         cl_device_id dev, 
                         const char* filename) {

   cl_program program;
   FILE *program_handle;
   char *program_buffer, *program_log;
   size_t program_size, log_size;
   int err;

   /* Read program file and place content into buffer */
   program_handle = fopen(filename, "r");
   if(program_handle == NULL) {
      perror("Couldn't find the program file");
      exit(1);
   }
   fseek(program_handle, 0, SEEK_END);
   program_size = ftell(program_handle);
   rewind(program_handle);
   program_buffer = (char*)malloc(program_size + 1);
   program_buffer[program_size] = '\0';
   fread(program_buffer, sizeof(char), program_size, program_handle);
   fclose(program_handle);

   program = clCreateProgramWithSource(ctx, 1, 
      (const char**)&program_buffer, &program_size, &err);
   if(err < 0) {
      perror("Couldn't create the program");
      exit(1);
   }
   free(program_buffer);

   err = clBuildProgram(program, 0, NULL, NULL, NULL, NULL);
   if(err < 0) {

      clGetProgramBuildInfo(program, dev, CL_PROGRAM_BUILD_LOG, 
            0, NULL, &log_size);
      program_log = (char*) malloc(log_size + 1);
      program_log[log_size] = '\0';
      clGetProgramBuildInfo(program, dev, CL_PROGRAM_BUILD_LOG, 
            log_size + 1, program_log, NULL);
      printf("%s\n", program_log);
      free(program_log);
      exit(1);
   }

   return program;
}

int main() {

  unsigned long n = 100000L;
  size_t local_size = 16;
  cl_int num_groups = n/local_size;

  /* OpenCL structures */
  cl_device_id device;
  cl_context context;
  cl_program program;
  cl_kernel kernel;
  cl_command_queue queue;
  cl_int i, j, err;
  
  device  = create_device();
  context = create_context(device);
  queue   = create_queue(device, context);
  program = build_program(context, device, PROGRAM_FILE);
  
  double * data1 = (double *) malloc(n*sizeof(double));
  double * data2 = (double *) malloc(n*sizeof(double));
  double * sum   = (double *) malloc(n*sizeof(double));

  /* Initialize data */
  for(i=0; i<n; i++) {
    double x = (1.0*i)/n;
    data1[i] = sin(x)*sin(x);
    data2[i] = cos(x)*cos(x);
  }

  cl_mem input1, input2, output;
  
  input1 = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_USE_HOST_PTR, 
                          n * sizeof(double), data1, &err);
  if(err < 0) {
    perror("Couldn't create input1 buffer");
    exit(1);   
  };

  input2 = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_USE_HOST_PTR,
                          n * sizeof(double), data2, &err);
  if(err < 0) {
    perror("Couldn't create input2 buffer");
    exit(1);   
  };

  output = clCreateBuffer(context, CL_MEM_WRITE_ONLY | CL_MEM_USE_HOST_PTR,
                          n * sizeof(double), sum, &err);
  if(err < 0) {
    perror("Couldn't create output buffer");
    exit(1);   
  };

  err = clEnqueueWriteBuffer(queue, input1, 
    CL_TRUE, 0, n * sizeof(double), data1, 0, NULL, NULL);
  if(err < 0) {
    perror("Couldn't copy data1");
    exit(1);   
  };

  err = clEnqueueWriteBuffer(queue, input2, 
    CL_TRUE, 0, n * sizeof(double), data2, 0, NULL, NULL);
  if(err < 0) {
    perror("Couldn't copy data2");
    exit(1);   
  };

  /* Create a kernel */
  kernel = clCreateKernel(program, KERNEL_FUNC, &err);
  if(err < 0) {
    perror("Couldn't create a kernel");
    exit(1);
  };

  /* Create kernel arguments */
  err  = clSetKernelArg(kernel, 0, sizeof(cl_mem), &input1);
  err |= clSetKernelArg(kernel, 1, sizeof(cl_mem), &input2);
  err |= clSetKernelArg(kernel, 2, sizeof(cl_mem), &output);
  err |= clSetKernelArg(kernel, 3, sizeof(unsigned long), &n);
  if(err < 0) {
    perror("Couldn't create a kernel argument");
    exit(1);
  }

  err = clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &n, 
                               &local_size, 0, NULL, NULL); 
  if(err < 0) {
    perror("Couldn't enqueue the kernel");
    exit(1);
  }

  /* Read the kernel's output    */
  err = clEnqueueReadBuffer(queue, output, CL_TRUE, 0, 
                            n * sizeof(double), sum, 0, NULL, NULL);
  if(err < 0) {
    perror("Couldn't read the buffer");
    exit(1);
  }

  err = 0;
  for (i=0; i<n; i++)
    if (fabs(sum[i] - 1.0) > 1e-10) {
      fprintf(stderr, "error on sum[%d]", i);
      err = 1;
      break;
    }
  if (err == 0)
    printf("\ncheck results ok\n\n");

  /* Deallocate resources */
  clReleaseKernel(kernel);
  clReleaseMemObject(output);
  clReleaseMemObject(input1);
  clReleaseMemObject(input2);
  clReleaseCommandQueue(queue);
  clReleaseProgram(program);
  clReleaseContext(context);
  free(sum);
  free(data1);
  free(data2);

  return 0;
}
